#pragma once

#include <cstdint>
#include <map>
#include <string>

struct Vector2;
struct Color;
struct Rectangle;
struct Texture;
struct RenderTexture;
struct Camera2D;

#ifdef PANDA_INCLUDE_TYPE

struct Vector2 {
  float x;
  float y;
};

struct Color {
  unsigned char r;
  unsigned char g;
  unsigned char b;
  unsigned char a;
};

struct Rectangle {
  float x;
  float y;
  float width;
  float height;
}

struct Texture {
  unsigned int id;
  int width;
  int height;
  int mipmaps;
  int format;
}

struct RenderTexture {
  unsigned int id;
  Texture texture;
  Texture depth;
}

struct Camera2D {
  Vector2 offset;
  Vector2 target;
  float rotation;
  float zoom;
}

#endif

namespace panda
{
  class Scene {
  public:
    Scene() {}
    virtual ~Scene() {}
    virtual void Preload() {}
    virtual void Update() {}

    bool preloaded = false;
  };

  struct Sprite {
    int width;
    int height;
    int offsetX;
    int offsetY;
    int frameX;
    int frameY;
    int frameW;
    int frameH;
    float pivotX;
    float pivotY;
    bool isRotated;
  };

  class TextureMap {
  public:
    Texture *texture;
    std::map<const char*, Sprite> map;

    TextureMap() {}
    virtual ~TextureMap() {}
    virtual void Init() {}
  };

  struct Matrix {
    float a = 1.0f;
    float b = 0.0f;
    float c = 0.0f;
    float d = 1.0f;
    float tx = 0.0f;
    float ty = 0.0f;

    void identity();
    void translate(float dx, float dy);
    void rotate(float radian);
    void scale(float sx, float sy);

    void concat(Matrix m);
  };

  class Random {
		public:
			Random(uint64_t seed = 521288629);
			void SetSeed(uint64_t seed);

			uint64_t Next();
			double Next_UNI();
			double Range(double min, double max);
		
		private:
			uint64_t s[4];
	};

  class App {
  public:
    void CreateWindow(int w, int h, const char *str);

    void AppUpdate();
    void SetScene(Scene* scene);
    Scene* GetScene();
  
  private:
    Scene* _scene_current = nullptr;
  };

  // ############################################################################################

  // ### GUI
  #ifdef PANDA_INCLUDE_IMGUI
  void GuiInit();
  void GuiDestroy();
  void GuiBegin();
  void GuiEnd();
  void GuiThemeSetup();
  #endif

  // ### gridmap
  // Draw GridMap using cam, use this inside BeginMode2D
  void DrawGridMap(Texture texture, Camera2D cam);
  // This is quick version without cam, does not support zoom
  void DrawGridMapQuick(Texture texture, Vector2 offset);
  Texture LoadDefaultGridMap();

  // ### draw
  void DrawTextureTiled(Texture texture, Rectangle source, Rectangle dest, Vector2 origin, float rotation, float scale, Color tint);
  void DrawRenderTexture(const RenderTexture &buffer, const float &x, const float &y, const Color &color);

  // ## sprite
  void DrawSprite(Texture &texture, const panda::Sprite &sprite, const int &x, const int &y, const Color &color);
  void DrawSpriteV(Texture &texture, const panda::Sprite &sprite, const Vector2 &pos, const Color &color);
  void DrawSpriteEx(Texture &texture, const panda::Sprite &sprite, const Vector2 &pos, const Vector2 &scale, const float &rotation, bool flipX, bool flipY, const Color &color);

  void DrawTextureMap(const panda::TextureMap &atlas, const char* texture, const int &x, const int &y, const Color &color);
  void DrawTextureMapV(const panda::TextureMap &atlas, const char* texture, const Vector2 &pos, const Color &color);
  void DrawTextureMapEx(const panda::TextureMap &atlas, const char* texture, const Vector2 &pos, const Vector2 &scale, const float &rotation, bool flipX, bool flipY, const Color &color);

  // Vector2 GetPointInSprite(const Vector2 &point, const panda::Sprite &sprite, const Vector2 &pos, const Vector2 &scale, const float &rotation);
  
  // ### texture utils
  Rectangle GetTextureRect(const Texture &tex, bool flipX, bool flipY);

  // Euclidean Distance
  double distance(const Vector2 &p1, const Vector2 &p2);
  double distanceSqrt(const Vector2 &p1, const Vector2 &p2);

  // Manhattan Distance
  double distanceManhattan(const Vector2 &p1, const Vector2 &p2);
}
